/*
 *  auto-nng v1.7
 *
 *
 *  Copyright (c) 2006-2008 Jan Behrens
 *  Copyright (c) 2011 Public Software Group e. V., Berlin
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a
 *  copy of this software and associated documentation files (the "Software"),
 *  to deal in the Software without restriction, including without limitation
 *  the rights to use, copy, modify, merge, publish, distribute, sublicense,
 *  and/or sell copies of the Software, and to permit persons to whom the
 *  Software is furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 *  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *  DEALINGS IN THE SOFTWARE.
 *
 */

#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <limits.h>
#include <inttypes.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>
#include <stdio.h>

#define MAX_LINE_LEN 16384
#define DOUBLE_FORMAT "%+1.9e"

const char *banner_string =
  "==================================================\n" \
  "  auto-nng v1.7                                   \n" \
  "  Copyright (c) 2011 Public Software Group e. V.  \n" \
  "                                                  \n" \
  "  This software is EXPERIMENTAL                   \n" \
  "  and comes with ABSOLUTELY NO WARRANTY.          \n" \
  "                                                  \n" \
  "  web:   http://www.public-software-group.org/    \n" \
  "==================================================\n\n";

const char *usage_fmtstr =
  "Example usage: %s i1 i2 i3 I4 o5 o6 O7 < learning_data.csv > network.nn" \
  "\n\n" \
  "Assuming there is a file learning_data.csv containing the learning \n" \
  "data as comma seperated values, having binary inputs (represented as \n" \
  "real numbers from -1.0 to +1.0) in the first three columns, having a\n" \
  "continuous input in the fourth column, having two binary outputs in \n" \
  "the fifth and sixth column, and a continious output in the seventh \n" \
  "column.\n\n" \
  "In order to use a previously generated network, call: \n" \
  "%s run network.nn < input_data.csv > output_data.csv\n\n";

volatile bool termination_flag = false;

void termination_signal_handler(int s) {
  termination_flag = true;
}

#define mutation_rate_factor(r) (exp((r) / 256.0 - 4.0))

typedef struct neuronal_network_struct {
  int layer_count;
  int input_count;
  int output_count;
  int weight_count;
  int *layer_sizes;
  double *weights;
  double *weight_buffer;
  double *mutation_rates;
  double *mutation_impacts;
  double *input_values;
  double *value_buffer;
  double network_error;
} neuronal_network_t;

void get_network_properties(int *layer_sizes,
  int *layer_count_ptr,
  int *input_count_ptr,
  int *output_count_ptr,
  int *max_width_ptr,
  int *weight_count_ptr
) {
  int input_count;
  int output_count = 0;
  int max_width = 0;
  int weight_count = 0;
  int layer;
  input_count = layer_sizes[0];
  assert(input_count > 0);
  for (layer = 0; ; layer++) {
    int width;
    width = layer_sizes[layer];
    assert(width >= 0);
    if (width == 0) break;
    weight_count += (width + 1) * layer_sizes[layer+1];
    output_count = width;
    if (width > max_width) max_width = width;
  }
  assert(layer >= 2);
  if (layer_count_ptr  != NULL) *layer_count_ptr  = layer;
  if (input_count_ptr  != NULL) *input_count_ptr  = input_count;
  if (output_count_ptr != NULL) *output_count_ptr = output_count;
  if (max_width_ptr    != NULL) *max_width_ptr    = max_width;
  if (weight_count_ptr != NULL) *weight_count_ptr = weight_count;
}

neuronal_network_t *create_network(int *layer_sizes) {
  int layer_count;
  int input_count;
  int output_count;
  int max_width;
  int weight_count;
  neuronal_network_t *result;
  get_network_properties(layer_sizes, &layer_count, &input_count, &output_count, &max_width, &weight_count);
  result = malloc(sizeof(neuronal_network_t) +
    (layer_count+1)  * sizeof(int) +
    (weight_count*4) * sizeof(double) +
    (max_width*2)    * sizeof(double));
  if (result == NULL) abort();
  result -> layer_sizes      = (int *)   ( (void *)(result)                     + sizeof(neuronal_network_t)       );
  result -> weights          = (double *)( (void *)(result -> layer_sizes)      + (layer_count+1) * sizeof(int)    );
  result -> weight_buffer    = (double *)( (void *)(result -> weights)          + weight_count    * sizeof(double) );
  result -> mutation_rates   = (double *)( (void *)(result -> weight_buffer)    + weight_count    * sizeof(double) );
  result -> mutation_impacts = (double *)( (void *)(result -> mutation_rates)   + weight_count    * sizeof(double) );
  result -> input_values     = (double *)( (void *)(result -> mutation_impacts) + weight_count    * sizeof(double) );
  result -> value_buffer     = (double *)( (void *)(result -> input_values)     + max_width       * sizeof(double) );
  result -> layer_count =  layer_count;
  result -> input_count =  input_count;
  result -> output_count = output_count;
  result -> weight_count = weight_count;
  bcopy(layer_sizes, result -> layer_sizes, (layer_count + 1) * sizeof(int));
  bzero(result -> weights, weight_count * sizeof(double));
  bzero(result -> mutation_rates, weight_count * sizeof(double));
  result -> network_error = -1.0;
  return result;
}

#define sat(x) ( \
  ((x) < -1e20) ? -1 : \
  ((x) >  1e20) ?  1 : \
  ((x) > -1e-5 && (x) < 1e-5) ? (x) : \
  ((x) < 0) ? ( -0.5/(x) - sqrt(0.25/((x)*(x)) + 1) ) : \
              ( -0.5/(x) + sqrt(0.25/((x)*(x)) + 1) ) )

#define unsat(x) ( \
  ((x) <= -1) ? (-1.0/0.0) : \
  ((x) >=  1) ? ( 1.0/0.0) : \
  ( (x) / (1 - (x)*(x)) ) )

double *outputs_calculation(neuronal_network_t *network) {
  int *layer_sizes;
  double *weights;
  double *inputs;
  double *buffer;
  int layer;
  int input_pos;
  int output_pos;
  int weight_pos = 0;
  layer_sizes = network -> layer_sizes;
  weights     = network -> weights;
  inputs      = network -> input_values;
  buffer      = network -> value_buffer;
  for (layer = 0; ; layer++) {
    int input_count;
    int output_count;
    input_count = layer_sizes[layer];
    output_count = layer_sizes[layer+1];
    assert(input_count > 0 && output_count >= 0);
    if (output_count == 0) break;
    for (output_pos = 0; output_pos < output_count; output_pos++) {
      double sum;
      sum = weights[weight_pos++];
      for (input_pos = 0; input_pos < input_count; input_pos++) {
        sum += weights[weight_pos++] * inputs[input_pos];
      }
      buffer[output_pos] = sat(sum);
    }
    { double *tmp; tmp = inputs; inputs = buffer; buffer = tmp; }
  }
  return inputs;
}

double network_error(neuronal_network_t *network, double *input_sets, double *output_sets, int set_count) {
  int input_count;
  int output_count;
  int set_position;
  int output_pos;
  double *inputs;
  double sum = 0.0;
  input_count  = network -> input_count;
  output_count = network -> output_count;
  inputs       = network -> input_values;
  assert(input_count > 0 && output_count > 0);
  for (set_position = 0; set_position < set_count; set_position++) {
    double *req_outputs;
    double *calc_outputs;
    req_outputs = output_sets + set_position * output_count;
    bcopy(input_sets + set_position * input_count, inputs, input_count * sizeof(double));
    calc_outputs = outputs_calculation(network);
    for (output_pos = 0; output_pos < output_count; output_pos++) {
      double error_value;
      error_value = calc_outputs[output_pos] - req_outputs[output_pos];
      sum += error_value * error_value;
    }
  }
  return sum / set_count;
}

double std_dist() {
  double rand1, rand2;
  rand1 = (random() & 0xFFFFFFFF) / 2147483648.0;
  rand2 = (random() & 0xFFFFFFFF) / 2147483648.0;
  return sqrt(-2.0 * log(1.0 - rand1)) * cos(2 * M_PI * rand2);
}

bool optimize_weights(neuronal_network_t *network, double *input_sets, double *output_sets, int set_count) {
  int weight_count;
  int weight_pos;
  double *old_weights;
  double old_error;
  double *new_weights;
  double new_error;
  double *mutation_rates;
  double *mutation_impacts;
  weight_count     = network -> weight_count;
  old_weights      = network -> weights;
  old_error        = network -> network_error;
  new_weights      = network -> weight_buffer;
  mutation_rates   = network -> mutation_rates;
  mutation_impacts = network -> mutation_impacts;
  if (old_error < 0.0) {
    old_error = network -> network_error = network_error(network, input_sets, output_sets, set_count);
  }
  for (weight_pos = 0; weight_pos < weight_count; weight_pos++) {
    double random_value;
    random_value = std_dist();
    new_weights[weight_pos] = (
      old_weights[weight_pos] +
      random_value * mutation_rate_factor(mutation_rates[weight_pos])
    );
    mutation_impacts[weight_pos] = random_value * random_value;
  }
  network -> weights = new_weights;
  new_error = network_error(network, input_sets, output_sets, set_count);
  if (new_error < old_error) {
    for (weight_pos = 0; weight_pos < weight_count; weight_pos++) {
      mutation_rates[weight_pos] += 2.0 * mutation_impacts[weight_pos];
    }
    network -> weight_buffer = old_weights;
    network -> network_error = new_error;
    return true;
  } else {
    for (weight_pos = 0; weight_pos < weight_count; weight_pos++) {
      mutation_rates[weight_pos] -= 0.5 * mutation_impacts[weight_pos];
    }
    network -> weights = old_weights;
    return false;
  }
}

void print_layer_sizes(neuronal_network_t *network) {
  int *layer_sizes;
  int layer_size;
  int layer_pos = 0;
  layer_sizes = network -> layer_sizes;
  while ((layer_size = layer_sizes[layer_pos++])) {
    fprintf(stderr, " %i", layer_size);
  }
}

neuronal_network_t *find_best_network(int input_count, int output_count,
  double *training_input, double *training_output, int training_count,
  double *test_input, double *test_output, int test_count
) {
  double d;
  int s1, s2, s3;
  int network_pos = 0;
  const int network_count = 7;
  int layer_sizes[5] = { 0, 0, 0, 0, 0 };
  neuronal_network_t **networks;
  neuronal_network_t *current_network;
  neuronal_network_t *best_network;
  double current_error;
  double best_error = -1.0;
  networks = malloc(network_count * sizeof(neuronal_network_t *));
  if (networks == NULL) abort();
  d = log(input_count*output_count) / M_LN2;
  s1 = round(d);
  s2 = round(2*d);
  s3 = round(3*d);
  if (s1 < 2) s1 = 2;
  if (s2 < s1+1) s2 = s1+1;
  if (s3 < s2+1) s3 = s2+1;
  layer_sizes[0] = input_count;
  layer_sizes[1] = output_count;
  networks[network_pos++] = create_network(layer_sizes);
  layer_sizes[2] = output_count;
  layer_sizes[1] = s1;
  networks[network_pos++] = create_network(layer_sizes);
  layer_sizes[1] = s2;
  networks[network_pos++] = create_network(layer_sizes);
  layer_sizes[1] = s3;
  networks[network_pos++] = create_network(layer_sizes);
  layer_sizes[3] = output_count;
  layer_sizes[1] = s1;
  layer_sizes[2] = s1;
  networks[network_pos++] = create_network(layer_sizes);
  layer_sizes[1] = round((double)(s3 * input_count + s1 * output_count) / (input_count+output_count));
  layer_sizes[2] = round((double)(s1 * input_count + s3 * output_count) / (input_count+output_count));
  if (layer_sizes[1] < 2) layer_sizes[1] = 2;
  if (layer_sizes[2] < 2) layer_sizes[2] = 2;
  if (layer_sizes[1] == s1 && layer_sizes[2] == s1) {
    if (input_count >= output_count) layer_sizes[1]++;
    if (output_count >= input_count) layer_sizes[2]++;
  } else if (layer_sizes[1] == s3 && layer_sizes[2] == s3) {
    if (input_count >= output_count) layer_sizes[2]--;
    if (output_count >= input_count) layer_sizes[1]--;
  }
  networks[network_pos++] = create_network(layer_sizes);
  layer_sizes[1] = s3;
  layer_sizes[2] = s3;
  networks[network_pos++] = create_network(layer_sizes);
  best_network = networks[0];
  do {
    for (network_pos = 0; network_pos < network_count; network_pos++) {
      current_network = networks[network_pos];
      if (optimize_weights(current_network, training_input, training_output, training_count)) {
        current_error = network_error(current_network, test_input, test_output, test_count);
        if ((best_error < 0) || (current_error < best_error)) {
          best_network = current_network;
          best_error = current_error;
          fprintf(stderr, "Training error: %f / Test error: %f / Layer sizes:", best_network -> network_error, best_error);
          print_layer_sizes(best_network);
          fprintf(stderr, "\n");
        }
      }
    }
  } while (!termination_flag);
  return best_network;
}

#define input_high_indicator(x)  ( sat( (x) - 1 )                           )
#define input_low_indicator(x)   ( input_high_indicator( -(x) )             )
#define output_high_indicator(x) ( input_high_indicator( unsat((x)/2) * 2 ) )
#define output_low_indicator(x)  ( input_low_indicator ( unsat((x)/2) * 2 ) )

#define read_indicators(l, h) ( \
  ( ((h) < -1+1e-15) ? -1 : ((h) > 1-1e-15) ? 1 : sat((unsat(h)+1) / 2) ) - \
  ( ((l) < -1+1e-15) ? -1 : ((l) > 1-1e-15) ? 1 : sat((unsat(l)+1) / 2) ) )

typedef struct dataset_struct {
  struct dataset_struct *next;
  double *binary_input_values;
  double *continuous_input_values;
  double *binary_output_values;
  double *continuous_output_values;
} dataset_t;

void io_error() {
  fprintf(stderr, "IO-Error: %s\n", strerror(errno));
  exit(2);
}

void data_parse_error() {
  fprintf(stderr, "Could not parse data.\n");
  exit(3);
}

void learning_data_insufficient() {
  fprintf(stderr, "Too few learning data supplied.\n");
  exit(4);
}

void network_parse_error() {
  fprintf(stderr, "Could not parse network data.\n");
  exit(5);
}

void main_generate(int argc, char **argv) {
  int binary_input_count = 0;
  int *binary_input_cols;
  int continuous_input_count = 0;
  int *continuous_input_cols;
  double *continuous_input_averages;
  double *continuous_input_stddevs;
  int binary_output_count = 0;
  int *binary_output_cols;
  int continuous_output_count = 0;
  int *continuous_output_cols;
  double *continuous_output_averages;
  double *continuous_output_stddevs;
  int column_count = 0;
  int row_count = 0;
  int input_set_length;
  int output_set_length;
  dataset_t *dataset = NULL;
  double *input_values;
  double *output_values;
  int timeout = 0;
  neuronal_network_t *network;
  srandom(time(NULL));
  if (argc > SIZE_MAX / sizeof(int)) abort();
  if (argc > 1) {
    if ((binary_input_cols      = malloc((argc - 1) * sizeof(int))) == NULL) abort();
    if ((continuous_input_cols  = malloc((argc - 1) * sizeof(int))) == NULL) abort();
    if ((binary_output_cols     = malloc((argc - 1) * sizeof(int))) == NULL) abort();
    if ((continuous_output_cols = malloc((argc - 1) * sizeof(int))) == NULL) abort();
  }
  {
    int i;
    for (i = 1; i < argc; i++) {
      char mode;
      char *number_str;
      char *number_endptr;
      long number;
      mode = argv[i][0];
      if (mode == '-') {
        mode = argv[i][1];
        if (mode != 't') goto command_line_error;
        number_str = argv[i] + 2;
        if (number_str[0] == 0) {
          if (++i >= argc) goto command_line_error;
          number_str = argv[i];
        }
        errno = 0;
        number = strtol(number_str, &number_endptr, 10);
        if (errno != 0 || number_endptr[0] != 0 || number < INT_MIN || number > INT_MAX) goto command_line_error;
        timeout = number;
      } else {
        if (mode != 'i' && mode != 'I' && mode != 'o' && mode != 'O') goto command_line_error;
        number_str = argv[i] + 1;
        if (number_str[0] == 0) goto command_line_error;
        errno = 0;
        number = strtol(number_str, &number_endptr, 10);
        if (errno != 0 || number_endptr[0] != 0 || number < INT_MIN || number > INT_MAX) goto command_line_error;
        if (number > column_count) column_count = number;
        number -= 1;
        if (number < 0) goto command_line_error;
        switch (mode) {
          case 'i': binary_input_cols     [binary_input_count++]      = number; break;
          case 'I': continuous_input_cols [continuous_input_count++]  = number; break;
          case 'o': binary_output_cols    [binary_output_count++]     = number; break;
          case 'O': continuous_output_cols[continuous_output_count++] = number; break;
        }
      }
    }
    if (binary_input_count == 0 && continuous_input_count == 0) goto command_line_error;
    if (binary_output_count == 0 && continuous_output_count == 0) goto command_line_error;
    goto no_command_line_error;
    command_line_error:
    if (argc >= 1) fprintf(stderr, usage_fmtstr, argv[0], argv[0]);
    exit(1);
    no_command_line_error: ;
  }
  {
    char line[MAX_LINE_LEN+2];
    double *value_buffer;
    value_buffer = malloc(column_count * sizeof(double));
    while (fgets(line, MAX_LINE_LEN+2, stdin)) {
      {
        char *str_pos;
        int column;
        str_pos = line;
        if (strlen(line) > MAX_LINE_LEN) data_parse_error();
        while (str_pos[0] == ' ' || str_pos[0] == '\r' || str_pos[0] == '\n' || str_pos[0] == '\t') str_pos++;
        if (str_pos[0] == 0) continue;
        for (column = 0; column < column_count; column++) {
          char *old_str_pos;
          errno = 0;
          old_str_pos = str_pos;
          value_buffer[column] = strtod(str_pos, &str_pos);
          if (errno != 0 || str_pos == old_str_pos) data_parse_error();
          while (str_pos[0] == ' ' || str_pos[0] == '\t' || str_pos[0] == ',' || str_pos[0] == ';') str_pos++;
        }
      }
      if (binary_input_count > SIZE_MAX/sizeof(double)/8) abort();
      if (continuous_input_count > SIZE_MAX/sizeof(double)/16) abort();
      if (binary_output_count > SIZE_MAX/sizeof(double)/8) abort();
      if (continuous_output_count > SIZE_MAX/sizeof(double)/16) abort();
      {
        dataset_t *new_dataset;
        new_dataset = malloc(
          sizeof(dataset_t) +
          (binary_input_count + continuous_input_count + binary_output_count + continuous_output_count) * sizeof(double));
        if (new_dataset == NULL) abort();
        if (dataset == NULL) {
          dataset = new_dataset;
          dataset -> next = dataset;
        } else {
          dataset_t *previous_dataset;
          previous_dataset = dataset;
          dataset = new_dataset;
          dataset -> next = previous_dataset -> next;
          previous_dataset -> next = dataset;
        }
      }
      dataset -> binary_input_values      = (double *)((void *)dataset + sizeof(dataset_t));
      dataset -> continuous_input_values  = (double *)((void *)(dataset -> binary_input_values)     + binary_input_count     * sizeof(double));
      dataset -> binary_output_values     = (double *)((void *)(dataset -> continuous_input_values) + continuous_input_count * sizeof(double));
      dataset -> continuous_output_values = (double *)((void *)(dataset -> binary_output_values)    + binary_output_count    * sizeof(double));
      {
        int i;
        for (i = 0; i < binary_input_count; i++)      dataset -> binary_input_values[i]      = value_buffer[binary_input_cols[i]];
        for (i = 0; i < continuous_input_count; i++)  dataset -> continuous_input_values[i]  = value_buffer[continuous_input_cols[i]];
        for (i = 0; i < binary_output_count; i++)     dataset -> binary_output_values[i]     = value_buffer[binary_output_cols[i]];
        for (i = 0; i < continuous_output_count; i++) dataset -> continuous_output_values[i] = value_buffer[continuous_output_cols[i]];
      }
      row_count++;
      if (row_count > INT_MAX / 2) abort();
    }
    if (ferror(stdin)) io_error();
    if (row_count < 2) learning_data_insufficient();
  }
  if (continuous_input_count != 0) {
    if (continuous_input_count > SIZE_MAX / sizeof(double)) abort();
    if ((continuous_input_averages = malloc(continuous_input_count * sizeof(double))) == NULL) abort();
    if ((continuous_input_stddevs  = malloc(continuous_input_count * sizeof(double))) == NULL) abort();
  }
  if (continuous_output_count != 0) {
    if (continuous_output_count > SIZE_MAX / sizeof(double)) abort();
    if ((continuous_output_averages = malloc(continuous_output_count * sizeof(double))) == NULL) abort();
    if ((continuous_output_stddevs  = malloc(continuous_output_count * sizeof(double))) == NULL) abort();
  }
  {
    int i;
    int row;
    double value;
    for (i = 0; i < continuous_input_count; i++) {
      double average = 0.0;
      double variance = 0.0;
      for (row = 0; row < row_count; row++) {
        dataset = dataset -> next;
        value = dataset -> continuous_input_values[i];
        average += value;
        variance += value * value;
      }
      average /= row_count;
      variance /= row_count;
      variance -= average * average;
      variance *= row_count;
      variance /= row_count - 1;
      continuous_input_averages[i] = average;
      continuous_input_stddevs[i] = sqrt(variance);
    }
    for (i = 0; i < continuous_output_count; i++) {
      double average = 0.0;
      double variance = 0.0;
      for (row = 0; row < row_count; row++) {
        dataset = dataset -> next;
        value = dataset -> continuous_output_values[i];
        average += value;
        variance += value * value;
      }
      average /= row_count;
      variance /= row_count;
      variance -= average * average;
      variance *= row_count;
      variance /= row_count - 1;
      continuous_output_averages[i] = average;
      continuous_output_stddevs[i] = sqrt(variance);
    }
  }
  input_set_length = binary_input_count + 2 * continuous_input_count;
  if (row_count > SIZE_MAX / input_set_length / sizeof(double)) abort();
  if ((input_values = malloc(row_count * input_set_length * sizeof(double))) == NULL) abort();
  output_set_length = binary_output_count + 2 * continuous_output_count;
  if (row_count > SIZE_MAX / output_set_length / sizeof(double)) abort();
  if ((output_values = malloc(row_count * output_set_length * sizeof(double))) == NULL) abort();
  {
    int row;
    int i;
    for (row = 0; row < row_count; row++) {
      {
        dataset_t *previous_dataset;
        int rows_to_skip;
        previous_dataset = dataset;
        dataset = dataset -> next;
        rows_to_skip = random() % (row_count - row);
        for (i = 0; i < rows_to_skip; i++) {
          previous_dataset = dataset;
          dataset = dataset -> next;
        }
        previous_dataset -> next = dataset -> next;
      }
      for (i = 0; i < binary_input_count; i++) {
        double value;
        value = dataset -> binary_input_values[i];
        if (value < -1.0) value = -1.0;
        else if (value > 1.0) value = 1.0;
        input_values[row * input_set_length + i] = value;
      }
      for (i = 0; i < continuous_input_count; i++) {
        double tmp;
        tmp = ((dataset -> continuous_input_values[i]) - continuous_input_averages[i]) / continuous_input_stddevs[i];
        input_values[row * input_set_length + binary_input_count + 2 * i] = input_low_indicator(tmp);
        input_values[row * input_set_length + binary_input_count + 2 * i + 1] = input_high_indicator(tmp);
      }
      for (i = 0; i < binary_output_count; i++) {
        double value;
        value = dataset -> binary_output_values[i];
        if (value < -1.0) value = -1.0;
        else if (value > 1.0) value = 1.0;
        output_values[row * output_set_length + i] = value;
      }
      for (i = 0; i < continuous_output_count; i++) {
        double tmp;
        tmp = ((dataset -> continuous_output_values[i]) - continuous_output_averages[i]) / continuous_output_stddevs[i];
        output_values[row * output_set_length + binary_output_count + 2 * i] = output_low_indicator(tmp);
        output_values[row * output_set_length + binary_output_count + 2 * i + 1] = output_high_indicator(tmp);
      }
      dataset = dataset -> next;  // important
    }
  }
  signal(SIGINT,  termination_signal_handler);
  signal(SIGALRM, termination_signal_handler);
  if (timeout > 0) alarm(timeout);
  network = find_best_network(input_set_length, output_set_length,
    input_values, output_values, row_count / 2,
    input_values + (row_count/2) * input_set_length, output_values + (row_count/2) * output_set_length, (row_count+1) / 2);
  fprintf(stderr, "\nFinishing...\n");
  {
    int i, j, k;
    int weight_pos = 0;
    printf("# auto-nng v1.6 neuronal network specification\n\n\n");
    printf("# number of inputs and outputs\n\n%i, %i\n\n\n", binary_input_count + continuous_input_count, binary_output_count + continuous_output_count);
    printf("# depth of neuronal network\n\n%i\n\n\n", network -> layer_count);
    printf("# neuronal network layer sizes\n\n");
    for (i = 0; i < network -> layer_count; i++) {
      if (i != 0) printf(", ");
      printf("%i", network -> layer_sizes[i]);
    }
    printf("\n\n\n");
    printf("# column numbers\n\n");
    if (binary_input_count != 0) {
      for (i = 0; i < binary_input_count; i++) {
        if (i != 0) printf(", ");
        printf("%i", binary_input_cols[i] + 1);
      }
      printf("\n");
    }
    if (continuous_input_count != 0) {
      for (i = 0; i < continuous_input_count; i++) {
        if (i != 0) printf(", ");
        printf("%i", continuous_input_cols[i] + 1);
      }
      printf("\n");
    }
    printf("\n");
    if (binary_output_count != 0) {
      for (i = 0; i < binary_output_count; i++) {
        if (i != 0) printf(", ");
        printf("%i", binary_output_cols[i] + 1);
      }
      printf("\n");
    }
    if (continuous_output_count != 0) {
      for (i = 0; i < continuous_output_count; i++) {
        if (i != 0) printf(", ");
        printf("%i", continuous_output_cols[i] + 1);
      }
      printf("\n");
    }
    printf("\n\n");
    if (continuous_input_count != 0 || continuous_output_count != 0) {
      printf("# transformation parameters\n\n");
      if (continuous_input_count != 0) {
        for (i = 0; i < continuous_input_count; i++) printf("%f, %f\n", continuous_input_averages[i], continuous_input_stddevs[i]);
      }
      if (continuous_input_count != 0 && continuous_output_count != 0) printf("\n");
      if (continuous_output_count != 0) {
        for (i = 0; i < continuous_output_count; i++) {
          printf(DOUBLE_FORMAT ", " DOUBLE_FORMAT "\n", continuous_output_averages[i], continuous_output_stddevs[i]);
        }
      }
      printf("\n\n");
    }
    printf("# weights\n\n");
    for (i = 0; i < network -> layer_count - 1; i++) {
      for (j = 0; j < (network -> layer_sizes)[i+1]; j++) {
        for (k = 0; k < (network -> layer_sizes)[i] + 1; k++) {
          if (k != 0) printf(", ");
          printf(DOUBLE_FORMAT, network -> weights[weight_pos++]);
        }
        printf("\n");
      }
      printf("\n");
    }
    printf("\n");
    printf("# end of file\n");
  }
}

void read_int_array(FILE *file, int *result_ptr, int length) {
  char line[MAX_LINE_LEN+2];
  int column;
  int value;
  char *str_pos;
  assert(length >= 0);
  if (length == 0) return;
  while (fgets(line, MAX_LINE_LEN+2, file)) {
    str_pos = line;
    if (strlen(line) > MAX_LINE_LEN) network_parse_error();
    while (str_pos[0] == ' ' || str_pos[0] == '\r' || str_pos[0] == '\n' || str_pos[0] == '\t') str_pos++;
    if (str_pos[0] == 0 || str_pos[0] == '#') continue;
    for (column = 0; column < length; column++) {
      char *old_str_pos;
      errno = 0;
      old_str_pos = str_pos;
      value = strtol(str_pos, &str_pos, 10);
      if (errno != 0 || str_pos == old_str_pos || value < INT_MIN || value > INT_MAX) network_parse_error();
      result_ptr[column] = value;
      while (str_pos[0] == ' ' || str_pos[0] == '\t' || str_pos[0] == ',' || str_pos[0] == ';') str_pos++;
    }
    if (str_pos[0] != 0 && str_pos[0] != '\r' && str_pos[0] != '\n') network_parse_error();
    break;
  }
  if (ferror(file)) io_error();
}

void read_double_array(FILE *file, double *result_ptr, int length) {
  char line[MAX_LINE_LEN+2];
  int column;
  double value;
  char *str_pos;
  assert(length >= 0);
  if (length == 0) return;
  while (fgets(line, MAX_LINE_LEN+2, file)) {
    str_pos = line;
    if (strlen(line) > MAX_LINE_LEN) network_parse_error();
    while (str_pos[0] == ' ' || str_pos[0] == '\r' || str_pos[0] == '\n' || str_pos[0] == '\t') str_pos++;
    if (str_pos[0] == 0 || str_pos[0] == '#') continue;
    for (column = 0; column < length; column++) {
      char *old_str_pos;
      errno = 0;
      old_str_pos = str_pos;
      value = strtod(str_pos, &str_pos);
      if (errno != 0 || str_pos == old_str_pos) network_parse_error();
      result_ptr[column] = value;
      while (str_pos[0] == ' ' || str_pos[0] == '\t' || str_pos[0] == ',' || str_pos[0] == ';') str_pos++;
    }
    if (str_pos[0] != 0 && str_pos[0] != '\r' && str_pos[0] != '\n') network_parse_error();
    break;
  }
  if (ferror(file)) io_error();
}

void main_run(char *filename) {
  FILE *file;
  int input_output_counts[2];
  int layer_count;
  int *layer_sizes;
  neuronal_network_t *network;
  int binary_input_count;
  int *binary_input_cols;
  int continuous_input_count;
  int *continuous_input_cols;
  double *continuous_input_stats;
  int binary_output_count;
  int *binary_output_cols;
  int continuous_output_count;
  int *continuous_output_cols;
  double *continuous_output_stats;
  double *values;
  double *input_values;
  double *output_values;
  int i, j;
  int input_column_count = 0;
  int column_count = 0;
  fprintf(stderr, "Loading neuronal network from file \"%s\".\n", filename);
  file = fopen(filename, "r");
  if (file == NULL) {
    fprintf(stderr, "Could not open file: %s\n", strerror(errno));
    exit(2);
  }
  read_int_array(file, input_output_counts, 2);
  read_int_array(file, &layer_count, 1);
  if (layer_count < 2) network_parse_error();
  if (layer_count > 10) network_parse_error();
  layer_sizes = malloc((layer_count + 1) * sizeof(int));
  if (layer_sizes == NULL) abort();
  layer_sizes[layer_count] = 0;
  read_int_array(file, layer_sizes, layer_count);
  network = create_network(layer_sizes);
  continuous_input_count = (network -> input_count) - input_output_counts[0];
  if (continuous_input_count < 0 || continuous_input_count > 256) network_parse_error();
  binary_input_count = input_output_counts[0] - continuous_input_count;
  if (binary_input_count < 0 || binary_input_count > 256) network_parse_error();
  continuous_output_count = (network -> output_count) - input_output_counts[1];
  if (continuous_output_count < 0 || continuous_output_count > 256) network_parse_error();
  binary_output_count = input_output_counts[1] - continuous_output_count;
  if (binary_output_count < 0 || binary_output_count > 256) network_parse_error();
  if (binary_input_count != 0      && (binary_input_cols      = malloc(binary_input_count      * sizeof(int))) == NULL) abort();
  if (continuous_input_count != 0  && (continuous_input_cols  = malloc(continuous_input_count  * sizeof(int))) == NULL) abort();
  if (binary_output_count != 0     && (binary_output_cols     = malloc(binary_output_count     * sizeof(int))) == NULL) abort();
  if (continuous_output_count != 0 && (continuous_output_cols = malloc(continuous_output_count * sizeof(int))) == NULL) abort();
  read_int_array(file, binary_input_cols     , binary_input_count);
  read_int_array(file, continuous_input_cols , continuous_input_count);
  read_int_array(file, binary_output_cols    , binary_output_count);
  read_int_array(file, continuous_output_cols, continuous_output_count);
  for (i = 0; i < binary_input_count; i++) binary_input_cols[i]--;
  for (i = 0; i < continuous_input_count; i++) continuous_input_cols[i]--;
  for (i = 0; i < binary_output_count; i++) binary_output_cols[i]--;
  for (i = 0; i < continuous_output_count; i++) continuous_output_cols[i]--;
  if (continuous_input_count != 0) {
    if ((continuous_input_stats = malloc(continuous_input_count * 2 * sizeof(double))) == NULL) abort();
    for (i = 0; i < continuous_input_count; i++) read_double_array(file, continuous_input_stats + 2 * i, 2);
  }
  if (continuous_output_count != 0) {
    if ((continuous_output_stats = malloc(continuous_output_count * 2 * sizeof(double))) == NULL) abort();
    for (i = 0; i < continuous_output_count; i++) read_double_array(file, continuous_output_stats + 2 * i, 2);
  }
  {
    int weight_pos = 0;
    int layer_input_count;
    int layer_output_count;
    double *weights;
    weights = network -> weights;
    for (i = 0; i < layer_count - 1; i++) {
      layer_input_count  = network -> layer_sizes[i] + 1;
      layer_output_count = network -> layer_sizes[i+1];
      for (j = 0; j < layer_output_count; j++) {
        read_double_array(file, weights + weight_pos, layer_input_count);
        weight_pos += layer_input_count;
      }
    }
    assert(weight_pos == (network -> weight_count));
  }
  fclose(file);
  fprintf(stderr, "Network loaded, processing data.\n");
  for (i = 0; i < binary_input_count;     i++) if (binary_input_cols[i]      + 1 > input_column_count) input_column_count = binary_input_cols[i]      + 1;
  for (i = 0; i < continuous_input_count; i++) if (continuous_input_cols[i]  + 1 > input_column_count) input_column_count = continuous_input_cols[i]  + 1;
  column_count = input_column_count;
  for (i = 0; i < binary_output_count;     i++) if (binary_output_cols[i]     + 1 > column_count) column_count = binary_output_cols[i]      + 1;
  for (i = 0; i < continuous_output_count; i++) if (continuous_output_cols[i] + 1 > column_count) column_count = continuous_output_cols[i]  + 1;
  values = malloc(column_count * sizeof(double));
  if (values == NULL) abort();
  {
    char line[MAX_LINE_LEN+2];
    bool first_value;
    input_values = network -> input_values;
    while (fgets(line, MAX_LINE_LEN+2, stdin)) {
      int column;
      {
        char *str_pos;
        str_pos = line;
        if (strlen(line) > MAX_LINE_LEN) data_parse_error();
        while (str_pos[0] == ' ' || str_pos[0] == '\r' || str_pos[0] == '\n' || str_pos[0] == '\t') str_pos++;
        if (str_pos[0] == 0) continue;
        for (column = 0; column < input_column_count; column++) {
          char *old_str_pos;
          errno = 0;
          old_str_pos = str_pos;
          values[column] = strtod(str_pos, &str_pos);
          if (errno != 0 || str_pos == old_str_pos) data_parse_error();
          while (str_pos[0] == ' ' || str_pos[0] == '\t' || str_pos[0] == ',' || str_pos[0] == ';') str_pos++;
        }
      }
      for (i = 0; i < binary_input_count; i++) input_values[i]  = values[binary_input_cols[i]];
      for (i = 0; i < continuous_input_count; i++) {
        double tmp;
        tmp = (values[continuous_input_cols[i]] - continuous_input_stats[2*i]) / continuous_input_stats[2*i+1];
        input_values[binary_input_count + 2 * i] = input_low_indicator(tmp);
        input_values[binary_input_count + 2 * i + 1] = input_high_indicator(tmp);
      }
      output_values = outputs_calculation(network);
      first_value = true;
      for (i = 0; i < column_count; i++) {
        for (j = 0; j < binary_input_count; j++) {
          if (binary_input_cols[j] == i) {
            if (first_value == false) printf(", ");
            printf(DOUBLE_FORMAT, values[i]);
            goto next_column;
          }
        }
        for (j = 0; j < continuous_input_count; j++) {
          if (continuous_input_cols[j] == i) {
            if (first_value == false) printf(", ");
            printf(DOUBLE_FORMAT, values[i]);
            goto next_column;
          }
        }
        for (j = 0; j < binary_output_count; j++) {
          if (binary_output_cols[j] == i) {
            if (first_value == false) printf(", ");
            printf(DOUBLE_FORMAT, output_values[j]);
            goto next_column;
          }
        }
        for (j = 0; j < continuous_output_count; j++) {
          if (continuous_output_cols[j] == i) {
            double low_indicator;
            double high_indicator;
            if (first_value == false) printf(", ");
            low_indicator = output_values[binary_output_count + 2 * j];
            high_indicator = output_values[binary_output_count + 2 * j + 1];
            printf(DOUBLE_FORMAT, read_indicators(low_indicator, high_indicator) * continuous_output_stats[2*j+1] + continuous_output_stats[2*j]);
            goto next_column;
          }
        }
        next_column:
        first_value = false;
      }
      printf("\n");
      fflush(stdout);
    }
  }
}

int main(int argc, char **argv) {
  fputs(banner_string, stderr);
  if (argc == 3 && !strcmp(argv[1], "run")) {
    main_run(argv[2]);
  } else {
    main_generate(argc, argv);
  }
  return 0;
}
